/*----------------------------------------------------------------------- 
 * 
 * File Name: CommTestMaster.c
 *
 * Author: Allen, B. and Creighton, J. D. E.
 * 
 * Revision: $Id: CommTestMaster.c,v 1.3 2000/07/23 01:12:49 jolien Exp $
 * 
 *-----------------------------------------------------------------------
 */

#include <stdio.h>
#include "LALStdlib.h"
#include "Comm.h"

NRCSID (COMMTESTMASTERC, "$Id: CommTestMaster.c,v 1.3 2000/07/23 01:12:49 jolien Exp $");

/* global variables */
#include "CommTestGlobal.h"

void
Master (LALStatus *status, MPIId id)
{
  MPIMessage   message;
  REAL4Vector *vector = NULL;
  INT4         numProcs;
  INT4         i;

  INITSTATUS (status, "Master", COMMTESTMASTERC);
  ATTATCHSTATUSPTR (status);

  printf ("Master starting up\n");

  LALSCreateVector (status->statusPtr, &vector, numPoints);
  CHECKSTATUSPTR (status);

  numProcs = id.numProcs;

  while (--numProcs > 0)
  {
    LALMPIRecvMsg (status->statusPtr, &message);
    CHECKSTATUSPTR (status);

    printf ("Master received message code %d from slave %d\n",
            message.msg, message.source);

    switch (message.msg)
    {
      case MPISVector:

        LALMPIRecvREAL4Vector (status->statusPtr, vector, message.source);
        CHECKSTATUSPTR (status);

        ASSERT (message.send == 1, status, 99, "Master expects to receive");

        printf ("Master received REAL4Vector from slave %d\n",
                message.source);

        for (i = 0; i < vector->length; ++i)
        {
          printf ("%f\t%f\n", vector->data[i], (REAL4)(i % 5 - 2));
        }

        break;

      default:

        printf ("Master: unexpected message code!\n");
        ++numProcs;
    }

  }

  LALSDestroyVector (status->statusPtr, &vector);
  CHECKSTATUSPTR (status);

  printf ("Master shutting down\n");

  DETATCHSTATUSPTR (status);
  RETURN (status);
}
